"""Data transfer object for estimated call object."""
from datetime import datetime, timedelta


class EstimatedCall:
    """Estimated call data."""

    def __init__(self, data: dict):
        """Initialize the estimated call object."""
        self._data = data

    @property
    def is_realtime(self) -> bool:
        """If the call is in real time."""
        return bool(self._data["realtime"])

    @property
    def aimed_arrival_time(self) -> datetime:
        """Aimed arrival time for the call according to the time table."""
        return datetime.strptime(self._data["aimedArrivalTime"], "%Y-%m-%dT%H:%M:%S%z")

    @property
    def expected_arrival_time(self) -> datetime:
        """Expected arrival time from realtime reports."""
        return datetime.strptime(
            self._data["expectedArrivalTime"], "%Y-%m-%dT%H:%M:%S%z"
        )

    @property
    def aimed_departure_time(self) -> datetime:
        """Aimed departure time for the call according to the time table."""
        return datetime.strptime(
            self._data["aimedDepartureTime"], "%Y-%m-%dT%H:%M:%S%z"
        )

    @property
    def expected_departure_time(self) -> datetime:
        """Expected departure time from realtime reports."""
        return datetime.strptime(
            self._data["expectedDepartureTime"], "%Y-%m-%dT%H:%M:%S%z"
        )

    @property
    def transport_mode(self) -> str:
        """Mode of transport."""
        return self._data["serviceJourney"]["journeyPattern"]["line"]["transportMode"]

    @property
    def delay_in_min(self) -> int:
        """Calculate delay in minutes."""
        return int(self.delay.total_seconds() / 60)

    @property
    def delay(self) -> timedelta:
        """Calculate delay."""
        return self.expected_departure_time - self.aimed_departure_time

    @property
    def line_id(self) -> str:
        """Id for the line driving the call."""
        return self._data["serviceJourney"]["journeyPattern"]["line"]["id"]

    @property
    def line_public_code(self) -> str:
        """Public id for the line driving the call."""
        return self._data["serviceJourney"]["journeyPattern"]["line"]["publicCode"]

    @property
    def front_display(self) -> str:
        """Text shown in front of the transport."""
        return (
            f"{self.line_public_code} {self._data['destinationDisplay']['frontText']}"
        )

    @property
    def raw(self):
        """Raw data for the place from the API."""
        return self._data



"""GraphQL query templates."""

GRAPHQL_STOP_TO_QUAY_QUERY = """
query(
    $stops: [String]!,
    $whitelist: InputWhiteListed,
    $omitNonBoarding: Boolean = true){
  stopPlaces(ids: $stops) {
    id
    quays(filterByInUse: true){
      id
      estimatedCalls(
          timeRange: 172100,
          numberOfDepartures: 1,
          whiteListed: $whitelist,
          omitNonBoarding: $omitNonBoarding){
        destinationDisplay {
          frontText
        }
      }
    }
  }
}
"""
GRAPHQL_STOP_TEMPLATE = """
  stopPlaces(ids: $stops) {
    id
    name
    estimatedCalls(
        whiteListed: $whitelist,
        numberOfDepartures: $numberOfDepartures) {
      ...callData
    }
  }
"""
GRAPHQL_QUAY_TEMPLATE = """
  quays(ids: $quays) {
    id
    name
    publicCode
    latitude
    longitude
    estimatedCalls(
        whiteListed: $whitelist,
        omitNonBoarding: $omitNonBoarding,
        numberOfDepartures: $numberOfDepartures) {
      ...callData
    }
  }
"""
GRAPHQL_CALL_FRAGMENT = """
fragment callData on EstimatedCall {
  realtime
  aimedArrivalTime
  aimedDepartureTime
  expectedArrivalTime
  expectedDepartureTime
  destinationDisplay {
    frontText
  }
  serviceJourney {
    journeyPattern {
      line {
        id
        transportMode
        publicCode
      }
    }
  }
}
"""
GRAPHQL_DEPARTURE_CALLS_QUERY = f"""
query(
  $stops: [String],
  $quays: [String],
  $whitelist: InputWhiteListed,
  $numberOfDepartures: Int = 2,
  $omitNonBoarding: Boolean = true){{
    {GRAPHQL_STOP_TEMPLATE}
    {GRAPHQL_QUAY_TEMPLATE}
}}
{GRAPHQL_CALL_FRAGMENT}
"""



"""Test for estimated calls object."""
from datetime import datetime, timedelta, timezone
from unittest import TestCase

from enturclient.dto import EstimatedCall


class EstimatedClassTests(TestCase):
    """Test case for dto EstimatedCall."""

    def setUp(self):
        """Set up test data and estimated call."""
        self._data = {
            "realtime": True,
            "aimedArrivalTime": "2020-03-15T21:57:00+0100",
            "aimedDepartureTime": "2020-03-15T21:58:00+0100",
            "expectedArrivalTime": "2020-03-15T22:56:00+0100",
            "expectedDepartureTime": "2020-03-15T22:59:05+0100",
            "destinationDisplay": {"frontText": "Voss"},
            "serviceJourney": {
                "journeyPattern": {
                    "line": {
                        "id": "NSB:Line:45",
                        "transportMode": "rail",
                        "publicCode": "45",
                    }
                }
            },
        }
        self._estimated_call = EstimatedCall(self._data)

    def test_parses_realtime_property(self):
        """Test if realtime is parsed correctly."""
        self.assertTrue(self._estimated_call.is_realtime)

    def test_parses_aimed_arrival_time(self):
        """Test if aimed arrival time is parsed correctly."""
        self.assertEqual(
            self._estimated_call.aimed_arrival_time,
            datetime(2020, 3, 15, 21, 57, 0, 0, timezone(timedelta(hours=1))),
        )

    def test_parses_expected_arrival_time(self):
        """Test if expected arrival time is parsed correctly."""
        self.assertEqual(
            self._estimated_call.expected_arrival_time,
            datetime(2020, 3, 15, 22, 56, 0, 0, timezone(timedelta(hours=1))),
        )

    def test_parses_aimed_departure_time(self):
        """Test if aimed departure time is parsed correctly."""
        self.assertEqual(
            self._estimated_call.aimed_departure_time,
            datetime(2020, 3, 15, 21, 58, 0, 0, timezone(timedelta(hours=1))),
        )

    def test_parses_expected_departure_time(self):
        """Test if expected departure time is parsed correctly."""
        self.assertEqual(
            self._estimated_call.expected_departure_time,
            datetime(2020, 3, 15, 22, 59, 5, 0, timezone(timedelta(hours=1))),
        )

    def test_parses_transport_mode(self):
        """Test if transport mode is parsed correctly."""
        self.assertEqual(self._estimated_call.transport_mode, "rail")

    def test_calculates_delay_in_min(self):
        """Test if EstimatedCall calculates delay in minutes correctly."""
        self.assertEqual(self._estimated_call.delay_in_min, 61)

    def test_calculates_delay(self):
        """Test if EstimatedCall calculates delay correctly."""
        self.assertEqual(self._estimated_call.delay, timedelta(seconds=3665))

    def test_parses_line_id(self):
        """Test if EstimatedCall parses input to give correct line id."""
        self.assertEqual(self._estimated_call.line_id, "NSB:Line:45")

    def test_parses_line_public_code(self):
        """Test if EstimatedCall parses input to give correct line public code."""
        self.assertEqual(self._estimated_call.line_public_code, "45")

    def test_parses_front_display(self):
        """Test if EstimatedCall parses input to give correct front display."""
        self.assertEqual(self._estimated_call.front_display, "45 Voss")

    def test_gives_input_out_as_raw(self):
        """Test if given input will be given out from raw property."""
        self.assertDictEqual(self._estimated_call.raw, self._data)



"""Test for place object."""
from unittest import TestCase

from enturclient.dto import Place


class PlaceTests(TestCase):
    """Test case for dto Place."""

    def setUp(self):
        """Set up test data and place object."""
        self._place_id = "NSR:StopPlace:548"
        self._place_name = "Bergen stasjon"
        self._place_estimated_calls = [{}, {}]
        self._data = {
            "id": self._place_id,
            "name": self._place_name,
            "estimatedCalls": self._place_estimated_calls,
        }
        self._place = Place(self._data, False)

    def test_parses_place_id(self):
        """Test if place id is parsed correctly."""
        self.assertEqual(self._place.place_id, self._place_id)

    def test_parses_name(self):
        """Test if name is parsed correctly."""
        self.assertEqual(self._place.name, self._place_name)

    def test_creates_estimated_calls(self):
        """Test if number of estimated calls are created."""
        self.assertEqual(
            len(self._place.estimated_calls), len(self._place_estimated_calls)
        )

    def test_parses_latitude(self):
        """Test if latitude is parsed correctly."""
        self.assertEqual(self._place.latitude, None)

    def test_parses_longitude(self):
        """Test if longitude is parsed correctly."""
        self.assertEqual(self._place.longitude, None)

    def test_parses_public_code(self):
        """Test if public code is parsed correctly."""
        self.assertEqual(self._place.public_code, None)

    def test_gives_input_out_as_raw(self):
        """Test if given input will be given out from raw property."""
        self.assertDictEqual(self._place.raw, self._data)


class PlacePlatformTests(PlaceTests):
    """Test case for dto Place."""

    def setUp(self):
        """Set up test data and place object."""
        self._place_id = "NSR:Quay:51852"
        self._place_name = "Kokstad"
        self._place_estimated_calls = [{}, {}, {}]
        self._data = {
            "id": self._place_id,
            "name": self._place_name,
            "estimatedCalls": self._place_estimated_calls,
            "publicCode": "",
            "latitude": 60.293217,
            "longitude": 5.267429,
        }
        self._place = Place(self._data, True)

    def test_parses_name(self):
        """Test if name is parsed correctly."""
        self.assertEqual(self._place.name, "Kokstad Platform 51852")

    def test_parses_latitude(self):
        """Test if latitude is parsed correctly."""
        self.assertEqual(self._place.latitude, 60.293217)

    def test_parses_longitude(self):
        """Test if longitude is parsed correctly."""
        self.assertEqual(self._place.longitude, 5.267429)

    def test_parses_public_code(self):
        """Test if public code is parsed correctly."""
        self.assertEqual(self._place.public_code, "")


class PlacePlatformWithPublicCodeTests(PlacePlatformTests):
    """Test case for dto Place."""

    def setUp(self):
        """Set up test data and place object."""
        super().setUp()
        self._data["publicCode"] = "123"
        self._place = Place(self._data, True)

    def test_parses_name(self):
        """Test if name is parsed correctly."""
        self.assertEqual(self._place.name, "Kokstad Platform 123")

    def test_parses_public_code(self):
        """Test if public code is parsed correctly."""
        self.assertEqual(self._place.public_code, "123")



"""Real-time information about public transport departures in Norway."""
import asyncio
import logging
from typing import Optional

import aiohttp

import enturclient.queries as q
from enturclient import dto

RESOURCE = "https://api.entur.io/journey-planner/v3/graphql"
_LOGGER = logging.getLogger(__name__)


class EnturPublicTransportData:
    """The Class for handling the data retrieval."""

    def __init__(
        self,
        client_name: str,
        stops: list = None,
        quays: list = None,
        line_whitelist: list = None,
        omit_non_boarding: bool = True,
        number_of_departures: int = 2,
        web_session: aiohttp.ClientSession = None,
    ):
        """Initialize the data object."""
        if not web_session:

            async def _create_session() -> aiohttp.ClientSession:
                return aiohttp.ClientSession()

            loop = asyncio.get_event_loop()
            self.web_session = loop.run_until_complete(_create_session())
        else:
            self.web_session = web_session

        self._client_name = client_name
        self._data: dict = {}
        self.stops = stops or []
        self.quays = quays or []
        self.omit_non_boarding = omit_non_boarding
        self.line_whitelist = line_whitelist or []
        self.number_of_departures = number_of_departures

        self.info: dict = {}

    async def close_connection(self):
        """Close the aiohttp session."""
        await self.web_session.close()

    def all_stop_places_quays(self) -> list:
        """Get all stop places and quays."""
        all_places = self.stops.copy()
        for quay in self.quays:
            all_places.append(quay)
        return all_places

    async def expand_all_quays(self) -> None:
        """Find all quays from stop places."""
        if not self.stops:
            return

        headers = {"ET-Client-Name": self._client_name}
        request = {
            "query": q.GRAPHQL_STOP_TO_QUAY_QUERY,
            "variables": {
                "stops": self.stops,
                "whitelist": {"lines": self.line_whitelist},
                "omitNonBoarding": self.omit_non_boarding,
            },
        }

        async with asyncio.timeout(15):
            async with self.web_session.post(
                RESOURCE, json=request, headers=headers
            ) as resp:
                if resp.status != 200:
                    _LOGGER.error(
                        "Error connecting to Entur, response http status code: %s",
                        resp.status,
                    )
                    return
                result = await resp.json()
        _LOGGER.debug("Got the following result from entur: %s", result)

        if "errors" in result:
            _LOGGER.warning(
                "Entur API responded with error message: %s", result["errors"]
            )
            return

        for stop_place in result["data"]["stopPlaces"]:
            if len(stop_place["quays"]) > 1:
                for quay in stop_place["quays"]:
                    if quay["estimatedCalls"]:
                        self.quays.append(quay["id"])

        _LOGGER.debug("Found the following queys after expansion: %s", self.quays)

    async def update(self) -> None:
        """Get the latest data from api.entur.org."""
        headers = {"ET-Client-Name": self._client_name}
        request = {
            "query": q.GRAPHQL_DEPARTURE_CALLS_QUERY,
            "variables": {
                "stops": self.stops,
                "quays": self.quays,
                "whitelist": {"lines": self.line_whitelist},
                "numberOfDepartures": self.number_of_departures,
                "omitNonBoarding": self.omit_non_boarding,
            },
        }

        async with asyncio.timeout(15):
            async with self.web_session.post(
                RESOURCE, json=request, headers=headers
            ) as resp:
                if resp.status != 200:
                    _LOGGER.error(
                        "Error connecting to Entur, response http status code: %s",
                        resp.status,
                    )
                    return
                result = await resp.json()
        _LOGGER.debug("Got the following result from entur: %s", result)

        if "errors" in result:
            _LOGGER.debug("Sent the following query to api: %s", request)
            _LOGGER.warning(
                "Entur API responded with error message: %s", result["errors"]
            )
            return

        self._data = result["data"]

        if "stopPlaces" in self._data:
            for stop in self._data["stopPlaces"]:
                self._process_place(stop, False)

        if "quays" in self._data:
            for quay in self._data["quays"]:
                self._process_place(quay, True)

        # find rate limits
        rate_limit_allowed = resp.headers.get("Rate-Limit-Allowed")
        rate_limit_available = resp.headers.get("Rate-Limit-Available")
        rate_limit_expiry_time = resp.headers.get("Rate-Limit-Expiry-Time")
        rate_limit_range = resp.headers.get("Rate-Limit-Range")
        rate_limit_used = resp.headers.get("Rate-Limit-Used")
        _LOGGER.debug(
            "Used %s of total %s in rate limits. Still %s available for use within time unit %s. Current rate limit expire at %s.",
            rate_limit_used,
            rate_limit_allowed,
            rate_limit_available,
            rate_limit_range,
            rate_limit_expiry_time,
        )

    def get_stop_info(self, stop_id: str) -> Optional[dto.Place]:
        """Get all information about a stop."""
        return self.info.get(stop_id)

    def _process_place(self, place: dict, is_platform: bool) -> None:
        """Extract information from place dictionary."""
        place_id = place["id"]
        self.info[place_id] = dto.Place(place, is_platform)



"""Data transfer object for Place object."""
from typing import Optional

from .estimated_call import EstimatedCall


class Place:
    """Represents a stop place or platform from entur."""

    def __init__(self, data: dict, is_platform: bool):
        """Initialize the place object."""
        self._data = data
        self.is_platform = is_platform

    @property
    def place_id(self) -> str:
        """Id for the stop place or platform."""
        return self._data["id"]

    @property
    def name(self) -> str:
        """Friendly name for the stop place or platform."""
        if self.is_platform:
            if self.public_code:
                return self._data["name"] + " Platform " + self.public_code
            return self._data["name"] + " Platform " + self.place_id.split(":")[-1]

        return self._data["name"]

    @property
    def latitude(self) -> Optional[float]:
        """Latitude part of place location."""
        return self._data.get("latitude")

    @property
    def longitude(self) -> Optional[float]:
        """Longitude part of place location."""
        return self._data.get("longitude")

    @property
    def public_code(self) -> Optional[int]:
        """Public code for the stop place."""
        return self._data.get("publicCode")

    @property
    def estimated_calls(self):
        """List estimated calls from the place."""
        return [EstimatedCall(s) for s in self._data["estimatedCalls"]]

    @property
    def raw(self):
        """Raw data for the place from the API."""
        return self._data
